/**
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * Copyright (C) 2013, 2014 Department of Informatics and
 * Telecommunications, National and Kapodistrian University of Athens
 *
 *      Author: Stella Giannakopoulou <sgian@di.uoa.gr>
 */

#include "method.h"
#include "utils.h"

extern char **interior_exterior_blocks;
extern int totalBlocks;

bool check_theorem(int maxnodeid)
{	
	for(int i = 0; i <= maxnodeid; i++)
	{
		int node1 = i;
		if(node_type[node1] == 'v')
		{
			//checking the first rule of the theorem
			if(variable_exterior_condition(node1) == false)
				return false;
				
			for(int j = 0; j <= maxnodeid; j++)
			{
				int node2 = j;
				//conditions between two variables
				if(node_type[node2] == 'v')
				{
					if(partial_overlap_conditions(node1, node2) == false)
						return false;
				}
				//condition between a variable and a landmark
				else if(node_type[node2] == 'l')
				{
					if(partial_overlap_conditions(node1, node2) == false)
						return false;
					if(proper_part_conditions(node1, node2) == false)
						return false;
				}
			}
		}
		//conditions for landmarks too
		//the first condition is omitted because it
		//comes of the arrangement
		else if(node_type[node1] == 'l')
		{
			for(int j = 0; j <= maxnodeid; j++)
			{
				int node2 = j;
				if(node_type[node2] == 'v')
				{
					if(partial_overlap_conditions(node1, node2) == false)
						return false;
				}
				//conditions between two landmarks
				else if(node_type[node2] == 'l')
				{
					if(partial_overlap_conditions(node1, node2) == false)
						return false;
					if(proper_part_conditions(node1, node2) == false)
						return false;
					#if defined(RCC5)	
					if(disjoint_condition(node1, node2) == false)
						return false;
					#endif		
				}
			}
		}
	}	
	return true;
}

//variable's exterior condition
//for any variable ui, E(ui) is a proper part of B
bool variable_exterior_condition(int var)
{
	int exterior_blocks = get_blocks_size(var, EXTERIOR);

	if(exterior_blocks == totalBlocks)
		return false;
	else
		return true;	
}

//for any V ui and L \/ V w  such that ui PO w, then:
//* E(ui) \/ E(w) is a proper part of B
//* E(ui) \/ I(w) is a proper part of B
//* I(ui) \/ E(w) is a proper part of B
bool partial_overlap_conditions(int node1, int node2)
{
	if(renz_matrix[node1][node2] == PO)
	{
		//construct the union of the exterior of node1 and the exterior of node2
		int exterior_union = get_union_blocks_size(node1, node2, EXTERIOR, EXTERIOR);
		if(exterior_union == totalBlocks)
			return false;

		//construct the union of the exterior of node1 and the interior of node2
		int exterior_interior_union = get_union_blocks_size(node1, node2, EXTERIOR, INTERIOR);
		
		if(exterior_interior_union == totalBlocks)				//the union of the exterior of node1 and the interior of node2 must not be equal to B
			return false;

		//construct the union of the interior of node1 and the exterior of node2
		int interior_exterior_union = get_union_blocks_size(node1, node2, INTERIOR, EXTERIOR);

		if(interior_exterior_union == totalBlocks)				//the union of the interior of node1 and the exterior of node2 must not be equal to B
			return false;
	}
	return true;
}

//for any V ui and L lj, if ui PP lj, then I(ui) is a proper part of I(lj)
//for any V ui and L lj, if lj PP ui, then E(ui) is a proper part of E(lj)
bool proper_part_conditions(int node1, int node2)
{	
	if(node_type[node2] == 'l')
	{
		#if defined(RCC5)
		if(renz_matrix[node1][node2] == (PP) || renz_matrix[node2][node1] == (PPI))
		#elif defined(RCC8)
		if(renz_matrix[node1][node2] == (TPP) || renz_matrix[node2][node1] == (TPPI))
		#endif
		{
			vector<int> interior1, interior2;
			vector<int>::iterator it1, it2;
			interior1 = get_blocks(node1, INTERIOR);
			interior2 = get_blocks(node2, INTERIOR);
			
			if(interior1.size() == interior2.size())
				return false;
			for(it1 = interior1.begin(); it1 != interior1.end(); it1++)
			{
				bool found = false;
				for(it2 = interior2.begin(); it2 != interior2.end(); it2++)
				{
					if((*it1) == (*it2))
					{
						found = true;
						break;
					}					
				}
				if(found == false)
					return false;
			}
		}
		#if defined(RCC5)
		else if(renz_matrix[node2][node1] == (PP) || renz_matrix[node1][node2] == (PPI))
		#elif defined(RCC8)
		else if(renz_matrix[node2][node1] == (TPP) || renz_matrix[node1][node2] == (TPPI))
		#endif
		{
			vector<int> exterior1, exterior2;
			vector<int>::iterator it1, it2;
			exterior1 = get_blocks(node1, EXTERIOR);
			exterior2 = get_blocks(node2, EXTERIOR);
			
			if(exterior1.size() == exterior2.size())
				return false;

			for(it1 = exterior1.begin(); it1 != exterior1.end(); it1++)
			{
				bool found = false;
				for(it2 = exterior2.begin(); it2 != exterior2.end(); it2++)
				{
					if((*it1) == (*it2))
					{
						found = true;
						break;
					}
				}
				if(found == false)
					return false;
			}
		}		
	}
	else
	{
		#if defined(RCC5)
		if(renz_matrix[node1][node2] == (PP) || renz_matrix[node1][node2] == (PPI))
		#elif defined(RCC8)
		if(renz_matrix[node1][node2] == (TPP) || renz_matrix[node1][node2] == (TPPI))
		#endif
		{
			//construct the union of the interior of node1 and the exterior of node2
			int interior_exterior_union = get_union_blocks_size(node1, node2, INTERIOR, EXTERIOR);
			
			if(interior_exterior_union == totalBlocks)				//the union of the interior of var1 and the exterior of var2 must not be equal to B
				return false;
		}		
	}
	return true;
}

#if defined(RCC5)
//for any L li and L lj  such that li DR lj, then:
//* I(li) is a proper part of E(lj)
bool disjoint_condition(int node1, int node2)
{
	if(renz_matrix[node1][node2] == DR)
	{
		vector<int> interior1 = get_blocks(node1, INTERIOR);
		vector<int> exterior2 = get_blocks(node2, EXTERIOR);
		vector<int>::iterator it1, it2;

		if(interior1.size() == exterior2.size()) //the sets cannot be the same, because of the
												 //existence of the unbounded block

				return false;

		for(it1 = interior1.begin(); it1 != interior1.end(); it1++)
		{
			bool found = false;
			for(it2 = exterior2.begin(); it2 != exterior2.end(); it2++)
			{
				if((*it1) == (*it2))
				{
					found = true;
					break;
				}
			}
			if(found == false)
				return false;
		}
	}
	return true;
}
#endif

/*returns how many interior or exterior blocks
 * a node contains */
int get_blocks_size(int node, char type)
{
	int size = 0;
	
	for(int i = 0; i < totalBlocks; i++)
	{
		if(interior_exterior_blocks[node][i] == type)
			size++;
	}
	return size;
}

/*returns the size of the union of blocks of two nodes
 * where type1, type 2 can be INTERIOR or EXTERIOR */
int get_union_blocks_size(int node1, int node2, char type1, char type2)
{
	int size = 0;
	
	for(int i = 0; i < totalBlocks; i++)
	{
		if(interior_exterior_blocks[node1][i] == type1 || interior_exterior_blocks[node2][i] == type2)
			size++;	
	}
	return size;
}

/*
 * return a vector with the requested (interior or
 * exterior) blocks
 * */
vector<int> get_blocks(int node, char type)
{
	vector<int> blocks;
	
	for(int i = 0; i < totalBlocks; i++)
	{
		if(interior_exterior_blocks[node][i] == type)
		{
			blocks.push_back(i);	
		}
	}
	return blocks;
}
