/**
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * Copyright (C) 2013, 2014 Department of Informatics and
 * Telecommunications, National and Kapodistrian University of Athens
 *
 *      Author: Stella Giannakopoulou <sgian@di.uoa.gr>
 */

#include "consolidated_data.h"

/************************Arrangement Construction**********************/

Consolidated_Data *d;

void Consolidated_Data::create_arrangement(ifstream &input_geometries)
{
	int numGeos = 0, numPolygons = 0, numMultiPolygons = 0, numPolygonPoints, numMultiPolygonPoints;
	string node;
	string wktGeometry;													//the string WKT literal of the current geometry read

	if(debug)
		cout<<"Creating arrangement..."<<endl;
	int i = 0;
	while(true)
	{
		input_geometries >> node;
		if(node[0] == '.')
			break;
		if(node[0] == '\n' || node[0] == '\0')
			continue;
		getline(input_geometries, wktGeometry);

		node_type[atoi(node.c_str())] = 'l';                     		//the node is a landmark
		WKTReader *reader = new WKTReader();
		Geometry *geom = reader->read(wktGeometry);
		if(!read_arrangement_from_file)
		{
			if(geom == NULL)
			{
				cerr << "consolidated_data.cpp: Error reading geometry" << endl;
				exit(1);
			}
			else
			{
				if(debug)
					cout << "Geometry: "<< i+1 <<endl;
				if(geom->getGeometryType() == "Polygon")
				{
					numPolygons++;
					Polygon* poly = dynamic_cast<Polygon*>(geom);				

					if(poly == NULL)
					{
						cerr << "consolidated_data.cpp: Error in dynamic cast" << endl;
						exit(1);
					}
					//cout<<"Number of holes: "<<poly->getNumInteriorRing()<<endl;
					numPolygonPoints = poly->getNumPoints();
					avgPolygonPoints += numPolygonPoints;

					if(numPolygons == 1)
					{
						minPolygonPoints = numPolygonPoints;
						maxPolygonPoints = numPolygonPoints;
					}
					else
					{
						if(numPolygonPoints < minPolygonPoints)
							minPolygonPoints = numPolygonPoints;

						if(numPolygonPoints > maxPolygonPoints)
							maxPolygonPoints = numPolygonPoints;
					}
					
					Polygon_2 pgn;
					CoordinateSequence *cords = poly->getExteriorRing()->getCoordinates();
					for (int c = 0; c < (cords->getSize() - 1); c++)
						pgn.push_back(Point_2((*cords)[c].x, (*cords)[c].y));
					
					insert_in_arrangement(cords, false, pgn.is_counterclockwise_oriented(), node);
					
					delete cords;
					for(int hole = 0; hole < poly->getNumInteriorRing(); hole++)
					{
						Polygon_2 pgn;
						CoordinateSequence *cords = poly->getInteriorRingN(hole)->getCoordinates();
						for (int c = 0; c < (cords->getSize() - 1); c++)
							pgn.push_back(Point_2((*cords)[c].x, (*cords)[c].y));

						insert_in_arrangement(cords, true, pgn.is_counterclockwise_oriented(), node);
						delete cords;
					}
				}
				else if(geom->getGeometryType() == "MultiPolygon")
				{
					numMultiPolygons++;
					
					MultiPolygon* multipoly = dynamic_cast<MultiPolygon*>(geom);
					if(multipoly == NULL)
					{
						cerr << "consolidated_data.cpp: Error in dynamic cast" << endl;
						exit(1);
					}
					numGeos = multipoly->getNumGeometries();
					avgGeometries+=numGeos;
					numMultiPolygonPoints = 0;
					for(int numGeo = 0; numGeo < multipoly->getNumGeometries(); numGeo++)
					{
						const Polygon *poly = dynamic_cast<const Polygon*>(multipoly->getGeometryN(numGeo));
						if(poly == NULL)
						{
							cerr << "consolidated_data.cpp: Error in dynamic cast" << endl;
							exit(1);
						}
						numMultiPolygonPoints += poly->getNumPoints();
						
						Polygon_2 pgn;
						CoordinateSequence *cords = poly->getExteriorRing()->getCoordinates();
						for (int c = 0; c < (cords->getSize() - 1); c++)
							pgn.push_back(Point_2((*cords)[c].x, (*cords)[c].y));
							
						insert_in_arrangement(cords, false, pgn.is_counterclockwise_oriented(), node);
						
						delete cords;
						for(int hole = 0; hole < poly->getNumInteriorRing(); hole++)
						{
							Polygon_2 pgn;
							CoordinateSequence *cords = poly->getInteriorRingN(hole)->getCoordinates();
							for (int c = 0; c < (cords->getSize() - 1); c++)
								pgn.push_back(Point_2((*cords)[c].x, (*cords)[c].y));
							
							insert_in_arrangement(cords, true, pgn.is_counterclockwise_oriented(), node);
							delete cords;
						}
					}
					avgMultiPolygonPoints+=numMultiPolygonPoints;
					if(numMultiPolygons == 1)
					{
						minMultiPolygonPoints = numMultiPolygonPoints;
						maxMultiPolygonPoints = numMultiPolygonPoints;

						minGeometries = numGeos;
						maxGeometries = numGeos;
					}
					else
					{
						if(numMultiPolygonPoints < minMultiPolygonPoints)
							minMultiPolygonPoints = numMultiPolygonPoints;

						if(numMultiPolygonPoints > maxMultiPolygonPoints)
							maxMultiPolygonPoints = numMultiPolygonPoints;

						if(numGeos < minGeometries)
							minGeometries = numGeos;
						if(numGeos > maxGeometries)
							maxGeometries = numGeos;
					}
				}
			}
		}
		i++;
		delete reader;
		delete geom;
	}
	totalGeometries = i;
	if(numPolygons > 0)
		avgPolygonPoints = (float) avgPolygonPoints/(float) numPolygons;
	if(numMultiPolygons > 0)
	{
		avgMultiPolygonPoints = (float) avgMultiPolygonPoints/(float) numMultiPolygons;
		avgGeometries = (float) avgGeometries/(float) numMultiPolygons;
	} 
}

void Consolidated_Data::insert_in_arrangement(CoordinateSequence *cords, bool is_hole, bool is_ccw, string numGeo)
{
	list<Data_Segment_2> data_segments;

	for (int i = 0; i < (cords->getSize() - 1); i++)
	{
		totalSegments++;
		//TODO: insert struct and not a string in the arrangement.
		ostringstream ostr;
		ostr << numGeo << " " << is_ccw << " " << is_hole << " " << Point_2((*cords)[i].x, (*cords)[i].y) << " " << Point_2 ((*cords)[i+1].x, (*cords)[i+1].y);
		data_segments.push_back(Data_Segment_2(Segment_2(Point_2((*cords)[i].x, (*cords)[i].y), Point_2 ((*cords)[i+1].x, (*cords)[i+1].y)), ostr.str()));
	}

	if(intersection == true)
		insert(arr, data_segments.begin(), data_segments.end());
	else
		insert_non_intersecting_curves(arr, data_segments.begin(), data_segments.end());
}

void Consolidated_Data::read_arrangement()
{	
	ifstream input_arrangement(arrangement_input_file);
	if(!input_arrangement.is_open())
	{
		cerr << "consolidated_data.cpp: Failed to open " << arrangement_input_file << " ..." <<endl;
		exit(1);
	}
	read(arr, input_arrangement, formatter);
	input_arrangement.close();
}

void Consolidated_Data::write_arrangement()
{
	//write arrangement to a file	
	ofstream output(arrangement_output_file);
	write(arr, output, formatter);
	output.close();
}

void Consolidated_Data::write_blocks()
{
	ofstream output(blocks_file);
	Cons_Arrangement_2::Face_const_iterator f;

	for(f = arr.faces_begin(); f != arr.faces_end(); ++f)
	{
		// Write the outer boundary of every bounded face.
		if(!f->is_unbounded())
		{
			output<<block_to_wkt(f->outer_ccb())->toString()<<endl;
		}
		// Print the boundary of each of the holes.
		/*Cons_Arrangement_2::Hole_const_iterator hi;
		int index = 1;
		for(hi = f->holes_begin(); hi != f->holes_end(); ++hi, ++index)
		{
			output << "    Hole #" << index << ": ";
			write_ccb (*hi, output);
		}*/
	}
	output.close();	
}

//convert a block in wkt format
Polygon* Consolidated_Data::block_to_wkt(const Cons_Arrangement_2::Ccb_halfedge_const_circulator circ)
{
	CoordinateArraySequence *points = new CoordinateArraySequence;
	Cons_Arrangement_2::Ccb_halfedge_const_circulator curr;
	curr = circ;
	Point_2 p = curr->source()->point();
	//to_double is needed if the kernel is exact
	points->add(Coordinate(to_double(p.x()), to_double(p.y())));
	//points->add(Coordinate(p.x(), p.y()));

	do
	{
		Cons_Arrangement_2::Halfedge_const_handle he = curr;
		p = he->target()->point();
		//to_double is needed if the kernel is exact
		points->add(Coordinate(to_double(p.x()), to_double(p.y())));
		//points->add(Coordinate(p.x(), p.y()));
	}
	while (++curr != circ);

	PrecisionModel *pm = new PrecisionModel();
	GeometryFactory *fac = new GeometryFactory(pm, srid);

	LinearRing *ring = new LinearRing(points, NULL);
	Polygon *poly = fac->createPolygon(ring, NULL);			//the blocks of the arrangement are only polygons without holes

	return poly;
}

void Consolidated_Data::arrangement_statistics()
{
	int blockPoints = 0, blockNum = 0;
	Cons_Arrangement_2::Face_const_iterator f;

	totalBlocks = arr.number_of_faces();
	for(f = arr.faces_begin(); f != arr.faces_end(); ++f)
	{
		if(!f->is_unbounded())
		{
			blockNum++;
			Cons_Arrangement_2::Ccb_halfedge_const_circulator circ = f->outer_ccb();
			Cons_Arrangement_2::Ccb_halfedge_const_circulator curr = circ;

			blockPoints = 1; //count the first point
			do
			{
				blockPoints++;
			} while (++curr != circ);
		}
		avgPointsPerBlock += blockPoints;
		if(blockNum == 1)
		{
			minPointsPerBlock = blockPoints;
			maxPointsPerBlock = blockPoints;
		}
		else
		{
			if(blockPoints < minPointsPerBlock)
				minPointsPerBlock = blockPoints;
			if(blockPoints > maxPointsPerBlock)
				maxPointsPerBlock = blockPoints;
		}
	}
	avgPointsPerBlock = (float) avgPointsPerBlock / (float) blockNum;
}

/************Computation of the interior/exterior blocks***************/


void Consolidated_Data::assign_blocks_to_landmarks(int geometries_start, int maxnodeid)
{
	Cons_Arrangement_2::Face_iterator f;
	string wktGeometry, node;
	
	if(debug)
		cout<<"Assigning blocks to polygons: "<<endl;
	//set the block number as data to each face
	int blockNum = 0;
	for(f = arr.faces_begin(); f != arr.faces_end(); ++f)
	{
		if(!f->is_unbounded())
		{
			blockNum++;
			f->set_data(blockNum);
		}
		else
			f->set_data(totalBlocks);
	}
	interior_exterior_blocks = new char*[maxnodeid + 1];
	for(int i = 0; i <= maxnodeid; i++)
		interior_exterior_blocks[i] = new char[totalBlocks];

	ifstream input_geometries(input_data);
	if(!input_geometries.is_open())
	{
		cerr << "consolidated_data.cpp: Failed to open " << input_data << " ..." <<endl;
		exit(1);
	}
	input_geometries.seekg(geometries_start, input_geometries.beg);
	
	if(parallel == true)
	{
		find_interior_exterior_blocks_threaded(input_geometries, totalGeometries, maxnodeid);
	}
	else
	{
		while(true)
		{
			input_geometries >> node;
			if(node[0] == '.')
				break;
			if(node[0] == '\n' || node[0] == '\0')
				continue;
			
			getline(input_geometries, wktGeometry);
			if(debug)
				cout<<"Checking Geometry "<< node <<endl;
			char *assigned_blocks = new char[totalBlocks];
			char *hole_blocks = new char[totalBlocks];
			find_interior_exterior_blocks(node, maxnodeid, assigned_blocks, hole_blocks);
			delete [] assigned_blocks;
			delete [] hole_blocks;
		}
	}
	input_geometries.close();
}

void Consolidated_Data::find_interior_exterior_blocks(string node, int maxnodeid, char *assigned_blocks, char *hole_blocks) const
{
	Cons_Arrangement_2::Face_const_iterator f;
	string is_hole;
	int node_no = atoi(node.c_str());
	int blockNum;
	
	memset(assigned_blocks, UNKNOWN, totalBlocks);
	memset(hole_blocks, '\0', totalBlocks);

	//iterate over all blocks
	for(f = arr.faces_begin(); f != arr.faces_end(); ++f)
	{
		if(!f->is_unbounded())
		{
			if(tangential_interior(f->outer_ccb(), node, is_hole, assigned_blocks, hole_blocks))
			{
				//if a face lies inside a polygon, then we can assume temporarily
				//that all the holes of the polygon will lie inside the polygon too.
				//This may change if we find out that these blocks are part of a hole
				//of the polygon
				assign_block_holes(f, node, is_hole, assigned_blocks, hole_blocks);
			}
			else
			{
				//the block is exterior because it is part of a hole of the polygon
				if(is_hole == TRUE)
					assign_block_holes(f, node, is_hole, assigned_blocks, hole_blocks);
			}
		}
		else
		{
			assigned_blocks[f->data()-1] = EXTERIOR;
			hole_blocks[f->data()-1] = NOT_HOLE;
		}
	}
	non_tangential_blocks(node, assigned_blocks, hole_blocks);
	
	for(f = arr.faces_begin(); f != arr.faces_end(); ++f)
	{
		if(!f->is_unbounded())
		{
			blockNum = f->data();
			
			if(assigned_blocks[blockNum - 1] == EXTERIOR)
			{
				if(debug)
				{
					cout<<"Block: "<<blockNum;
					cout<<" Exterior block"<<endl;
				}
				if(!parallel)
					interior_exterior_blocks[node_no][blockNum - 1] = EXTERIOR;
			}
			else if(assigned_blocks[blockNum - 1] == INTERIOR)
			{	
				if(debug)
				{
					cout<<"Block: "<<blockNum;
					cout<<" Interior block"<<endl;
				}
				if(!parallel)
					interior_exterior_blocks[node_no][blockNum - 1] = INTERIOR;
			}
		}
		//the unbounded block is always exterior
		else
		{
			if(debug)
			{
				cout<<"Block: "<<totalBlocks;
				cout<<" Exterior block"<<endl;
			}
			if(!parallel)
				interior_exterior_blocks[node_no][totalBlocks - 1] = EXTERIOR;
		}
	}
}

//return if the block with edge circulator `circ` is tangentially inside the landmark `node`
//return false if the block is tangentially outside the landmark
bool Consolidated_Data::tangential_interior(const Cons_Arrangement_2::Ccb_halfedge_const_circulator circ, string node, string &is_hole, char *assigned_blocks, char *hole_blocks) const
{
	//iterate over all the edges of the block to find
	//an edge that has been produced by the insertion
	//of the current `node` in the arrangement. This
	//means that this block is interior of the polygon
	//that represents the `node`
	Cons_Arrangement_2::Ccb_halfedge_const_circulator curr;
	curr = circ;

	do
	{
		Cons_Arrangement_2::Halfedge_const_handle he = curr;
		Data_Traits_2::Data_container::const_iterator dit;

		//the data of each edge is of the form: 'geometryNum is_ccw is_hole segmentStart segmentEnd'
		for (dit = he->curve().data().begin(); dit != he->curve().data().end(); ++dit)
		{
			istringstream data(*dit);
			string polygon, ccw, x1, y1, x2, y2;
			getline(data, polygon, ' ');
			if(polygon == node)
			{				
				getline(data, ccw, ' ');
				getline(data, is_hole, ' ');
				getline(data, x1, ' ');
				getline(data, y1, ' ');
				getline(data, x2, ' ');
				getline(data, y2, ' ');

				Point_2 source(atof(x1.c_str()), atof(y1.c_str()));
				Point_2 target(atof(x2.c_str()), atof(y2.c_str()));

				hole_blocks[he->face()->data() - 1] = is_hole[0];
		
				if(source != target && !(he->source()->point().x() == he->target()->point().x() && he->source()->point().y() == he->target()->point().y()))
				{
					if(ccw == TRUE)
					{
						if(same_direction(source, target, he->source()->point(), he->target()->point()))
						{						
							if(is_hole == FALSE)
							{
								assigned_blocks[he->face()->data() - 1] = INTERIOR;
								return true;
							}
							else
							{
								assigned_blocks[he->face()->data() - 1] = EXTERIOR;
								return false;
							}
						}
						else
						{
							if(is_hole == FALSE)
							{
								assigned_blocks[he->face()->data() - 1] = EXTERIOR;
								return false;
							}
							//we cannot decide the `else` case, because
							//the block may be interior if it belongs outside a hole
							//but exterior if it belongs inside another hole
						}
					}
					else
					{
						if(!same_direction(source, target, he->source()->point(), he->target()->point()))
						{
							if(is_hole == FALSE)
							{
								assigned_blocks[he->face()->data() - 1] = INTERIOR;
								return true;
							}
							else
							{
								assigned_blocks[he->face()->data() - 1] = EXTERIOR;
								return false;
							}
						}
						else
						{
							if(is_hole == FALSE)
							{
								assigned_blocks[he->face()->data() - 1] = EXTERIOR;
								return false;
							}
						}
					}
					break;
				}
			}
		}
	}
	while (++curr != circ);

	return false;
}

//this function deals with the holes of a block, in the cases where the block has been found
//to be interior of a polygon or exterior because it is part of a hole of the polygon
//if this block is interior block of a polygon, then all its holes (if they are not holes of the polygon) will be interior too
//similarly, if a block lies inside a hole of a polygon, then all its holes will be exterior too
void Consolidated_Data::assign_block_holes(const Cons_Arrangement_2::Face_const_handle &f, string node, string is_hole, char *assigned_blocks, char *hole_blocks) const
{
	Cons_Arrangement_2::Hole_const_iterator hi;

	for(hi = f->holes_begin(); hi != f->holes_end(); ++hi)
	{
		Cons_Arrangement_2::Ccb_halfedge_const_circulator circ, curr;
		circ = *hi;
		curr = circ;

		do
		{
			Cons_Arrangement_2::Halfedge_const_handle he = curr;
			//if the block lies inside a hole then it is definitely exterior
			if(is_hole == TRUE)
			{
				if(assigned_blocks[he->twin()->face()->data()-1] != EXTERIOR)
				{
					assigned_blocks[he->twin()->face()->data()-1] = EXTERIOR;
					hole_blocks[he->twin()->face()->data()-1] = is_hole[0];
					//set_hole_faces_exterior(he->twin()->face()->outer_ccb(), node);
					//continue using recursion to the hole's holes
					assign_block_holes(he->twin()->face(), node, is_hole, assigned_blocks, hole_blocks);
				}
			}
			//if it does not belong to a hole, then it is interior, but this may change
			else if (is_hole == FALSE)
			{
				if(assigned_blocks[he->twin()->face()->data()-1] == UNKNOWN)
				{
					assigned_blocks[he->twin()->face()->data()-1] = INTERIOR;
					hole_blocks[he->twin()->face()->data()-1] = is_hole[0];
					assign_block_holes(he->twin()->face(), node, is_hole, assigned_blocks, hole_blocks);
				}
			}
		}
		while (++curr != circ);
	}
}

void Consolidated_Data::set_hole_faces_exterior(const Cons_Arrangement_2::Ccb_halfedge_const_circulator circ, string node, char *assigned_blocks, char *hole_blocks)  const
{
	bool found;
	Cons_Arrangement_2::Ccb_halfedge_const_circulator curr = circ;

	do
	{
		Cons_Arrangement_2::Halfedge_const_handle he = curr;
		Data_Traits_2::Data_container::const_iterator dit;

		found = false;
		//the data of each edge is of the form: 'geometryNum is_ccw is_hole segmentStart segmentEnd'
		for (dit = he->curve().data().begin(); dit != he->curve().data().end(); ++dit)
		{
			istringstream data(*dit);
			string polygon;
			getline(data, polygon, ' ');
			if(polygon == node)
			{
				found=true;
				break;
			}
		}
		if(found == false)
		{
			if(assigned_blocks[he->twin()->face()->data() - 1] != EXTERIOR)
			{
				assigned_blocks[he->twin()->face()->data() - 1] = EXTERIOR;
				hole_blocks[he->twin()->face()->data() - 1] = HOLE;
			}
		}
	}
	while (++curr != circ);
}

//this function deals with blocks that have not been assigned yet
//as interior or exterior. These blocks are definitely non-tangential
//with the landmark
void Consolidated_Data::non_tangential_blocks(string node, char *assigned_blocks, char *hole_blocks) const
{
	Cons_Arrangement_2::Face_const_iterator f;
	int blockNum;
	bool unassignedLeft = false;

	for(f = arr.faces_begin(); f != arr.faces_end(); ++f)
	{
		if(!f->is_unbounded())
		{
			blockNum = f->data();

			if(assigned_blocks[blockNum - 1] == UNKNOWN)
			{
				if(non_tangential_interior(f->outer_ccb(), node, assigned_blocks, hole_blocks))
				{
					if(assigned_blocks[blockNum - 1] == INTERIOR)
						assign_block_holes(f, node, FALSE, assigned_blocks, hole_blocks);
				}
				if(assigned_blocks[blockNum - 1] == UNKNOWN)
				{
					unassignedLeft = true;
				}
			}
		}
	}
	//continue recursively
	if(unassignedLeft == true)
	{
		non_tangential_blocks(node, assigned_blocks, hole_blocks);
	}
}

//this function returns true if the block is non-tangential proper part of the landmark
//and false if the block is completely disconnected
bool Consolidated_Data::non_tangential_interior(const Cons_Arrangement_2::Ccb_halfedge_const_circulator circ, string node, char *assigned_blocks, char *hole_blocks) const
{
	Cons_Arrangement_2::Ccb_halfedge_const_circulator curr = circ;
	
	do
	{
		Cons_Arrangement_2::Halfedge_const_handle he = curr;
		if(assigned_blocks[he->twin()->face()->data() - 1] == EXTERIOR)
		{
			if(hole_blocks[he->twin()->face()->data() - 1] == NOT_HOLE)
			{
				assigned_blocks[circ->face()->data()-1] = EXTERIOR;
				hole_blocks[circ->face()->data()-1] = NOT_HOLE;
				return false;
			}
			//if the neighboring block is exterior because it is a hole
			//it is located inside the area of the polygon
			else
			{
				assigned_blocks[circ->face()->data()-1] = INTERIOR;
				hole_blocks[circ->face()->data()-1] = NOT_HOLE;
				return true;
			}
		}
		else if(assigned_blocks[he->twin()->face()->data() - 1] == INTERIOR)
		{
			assigned_blocks[circ->face()->data()-1] = INTERIOR;
			hole_blocks[circ->face()->data()-1] = NOT_HOLE;
			return true;
		}
	}
	while (++curr != circ);
	return false;
}


/**********************************Parallel****************************/
void block_landmark_position(string node, int maxnodeid, thread_result *res)
{
	int node_no = atoi(node.c_str());
	char *hole_blocks = new char[totalBlocks];
		
	res->node = node_no;
	d->find_interior_exterior_blocks(node, maxnodeid, res->blocks, hole_blocks);
	delete [] hole_blocks;
}

void Consolidated_Data::find_interior_exterior_blocks_threaded(ifstream &input_geometries, int totalGeometries, int maxnodeid)
{
	string node, wktGeometry;
	d = new Consolidated_Data;
	d->arr = arr;

	while(true)
	{
		input_geometries >> node;
		if(node[0] == '.')
			break;
		if(node[0] == '\n' || node[0] == '\0')
			continue;
			
		getline(input_geometries, wktGeometry);
		thread_result *res = new thread_result(); // initialize thread information
		tres.push_back(res);
		workers.push_back(new boost::thread(block_landmark_position, node, maxnodeid, res));
	}

	if(debug)
	{
		cout << "main thread: waiting for threads to finish...\n";
		cout << "main thread: you'd better have a look at htop\n";
	}

	// wait all threads to finish their work
	wait_threads(workers);

	// do something with the results
	// do something with the results
	for(list<thread_result*>::iterator it = tres.begin(); it != tres.end(); ++it) 
	{
		if(debug)
			cout<<"Checking Geometry "<<(*it)->node<<endl;
		for(int i = 0; i < totalBlocks; i++)
		{
			if(debug)
				cout<<"Block "<<i+1<<" "<<(*it)->blocks[i]<<endl;
			interior_exterior_blocks[(*it)->node][i] = (*it)->blocks[i];		
		}
	}
	// de-allocate memory
	free_threads(workers);
	free_thread_results(tres);
}

bool Consolidated_Data::check_validity(const Cons_Arrangement_2 &arr)
{
	Cons_Arrangement_2::Face_const_iterator f;

	for(f = arr.faces_begin(); f != arr.faces_end(); ++f)
	{
		if(!f->is_unbounded())
		{
			Polygon *poly = block_to_wkt(f->outer_ccb());
			if(poly->isValid() == false)
				return false;
		}
	}
	return true;
}

bool Consolidated_Data::is_valid(const Cons_Arrangement_2::Face_const_iterator &f)
{
       Polygon *poly = block_to_wkt(f->outer_ccb());
       if(poly->isValid() == false)
               return false;
       else
               return true;
}

bool same_direction(Point_2 source1, Point_2 target1, Point_2 source2, Point_2 target2)
{
	if((source1.x() < target1.x() && source2.x() < target2.x()) || (source1.x() > target1.x() && source2.x() > target2.x()))
	{
		return true;
	}
	if(source1.x() == target1.x())
	{
		if((source1.y() < target1.y() && source2.y() < target2.y()) || (source1.y() > target1.y() && source2.y() > target2.y()))
			return true;
	}
	return false;
}
