/**
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * Copyright (C) 2013, 2014 Department of Informatics and
 * Telecommunications, National and Kapodistrian University of Athens
 *
 *      Author: Stella Giannakopoulou <sgian@di.uoa.gr>
 */

#include "interior_point.h"

list<Point_2> shared_blockPoints;

/************************Arrangement Construction**********************/
void Interior_Point::create_arrangement(ifstream &input_geometries)
{
	int numGeos = 0, numPolygons = 0, numMultiPolygons = 0, numPolygonPoints, numMultiPolygonPoints;
	string node;
	string wktGeometry;													//the string WKT literal of the current geometry read

	if(debug)
		cout<<"Creating arrangement..."<<endl;
	int i = 0;
	while(true)
	{
		input_geometries >> node;
		if(node[0] == '.')
			break;
		if(node[0] == '\n' || node[0] == '\0')
			continue;
		getline(input_geometries, wktGeometry);

		node_type[atoi(node.c_str())] = 'l';                     		//the node is a landmark

		WKTReader *reader = new WKTReader();
		Geometry *geom = reader->read(wktGeometry);
		if(!read_arrangement_from_file)
		{
			if(geom == NULL)
			{
				cerr << "split.cpp: Error reading geometry" << endl;
				exit(1);
			}
			else
			{
				if(debug)
					cout << "Geometry: "<< i+1 <<endl;
				if(geom->getGeometryType() == "Polygon")
				{
					numPolygons++;
					Polygon* poly = dynamic_cast<Polygon*>(geom);				

					if(poly == NULL)
					{
						cerr << "split.cpp: Error in dynamic cast" << endl;
						exit(1);
					}
					//cout<<"Number of holes: "<<poly->getNumInteriorRing()<<endl;
					numPolygonPoints = poly->getNumPoints();
					avgPolygonPoints += numPolygonPoints;

					if(numPolygons == 1)
					{
						minPolygonPoints = numPolygonPoints;
						maxPolygonPoints = numPolygonPoints;
					}
					else
					{
						if(numPolygonPoints < minPolygonPoints)
							minPolygonPoints = numPolygonPoints;

						if(numPolygonPoints > maxPolygonPoints)
							maxPolygonPoints = numPolygonPoints;
					}
					
					//Polygon_2 pgn;
					CoordinateSequence *cords = poly->getExteriorRing()->getCoordinates();
					//for (int c = 0; c < (cords->getSize() - 1); c++)
					//	pgn.push_back(Point_2((*cords)[c].x, (*cords)[c].y));
					
					insert_in_arrangement(cords);
					delete cords;
					for(int hole = 0; hole < poly->getNumInteriorRing(); hole++)
					{
						//Polygon_2 pgn;
						CoordinateSequence *cords = poly->getInteriorRingN(hole)->getCoordinates();
						//for (int c = 0; c < (cords->getSize() - 1); c++)
						//	pgn.push_back(Point_2((*cords)[c].x, (*cords)[c].y));

						insert_in_arrangement(cords);
						delete cords;
					}
				}
				else if(geom->getGeometryType() == "MultiPolygon")
				{
					numMultiPolygons++;
					//cout<<"Type: "<<geom->getGeometryType()<<" and contains "<<geom->getNumGeometries() <<" geometries"<<endl;
					MultiPolygon* multipoly = dynamic_cast<MultiPolygon*>(geom);
					if(multipoly == NULL)
					{
						cerr << "interior_point.cpp: Error in dynamic cast" << endl;
						exit(1);
					}
					numGeos = multipoly->getNumGeometries();
					avgGeometries+=numGeos;
					numMultiPolygonPoints = 0;
					for(int numGeo = 0; numGeo < multipoly->getNumGeometries(); numGeo++)
					{
						const Polygon *poly = dynamic_cast<const Polygon*>(multipoly->getGeometryN(numGeo));
						if(poly == NULL)
						{
							cerr << "interior_point.cpp: Error in dynamic cast" << endl;
							exit(1);
						}
						numMultiPolygonPoints += poly->getNumPoints();
						
						//Polygon_2 pgn;
						CoordinateSequence *cords = poly->getExteriorRing()->getCoordinates();
						//for (int c = 0; c < (cords->getSize() - 1); c++)
						//	pgn.push_back(Point_2((*cords)[c].x, (*cords)[c].y));
							
						insert_in_arrangement(cords);
						delete cords;
						for(int hole = 0; hole < poly->getNumInteriorRing(); hole++)
						{
							//Polygon_2 pgn;
							CoordinateSequence *cords = poly->getInteriorRingN(hole)->getCoordinates();
							//for (int c = 0; c < (cords->getSize() - 1); c++)
							//	pgn.push_back(Point_2((*cords)[c].x, (*cords)[c].y));
							
							insert_in_arrangement(cords);
							delete cords;
						}
					}
					avgMultiPolygonPoints+=numMultiPolygonPoints;
					if(numMultiPolygons == 1)
					{
						minMultiPolygonPoints = numMultiPolygonPoints;
						maxMultiPolygonPoints = numMultiPolygonPoints;

						minGeometries = numGeos;
						maxGeometries = numGeos;
					}
					else
					{
						if(numMultiPolygonPoints < minMultiPolygonPoints)
							minMultiPolygonPoints = numMultiPolygonPoints;

						if(numMultiPolygonPoints > maxMultiPolygonPoints)
							maxMultiPolygonPoints = numMultiPolygonPoints;

						if(numGeos < minGeometries)
							minGeometries = numGeos;
						if(numGeos > maxGeometries)
							maxGeometries = numGeos;
					}
				}
			}
		}	
		i++;
		delete reader;
		delete geom;
	}
	if(numPolygons > 0)
		avgPolygonPoints = (float) avgPolygonPoints/(float) numPolygons;
	if(numMultiPolygons > 0)
	{
		avgMultiPolygonPoints = (float) avgMultiPolygonPoints/(float) numMultiPolygons;
		avgGeometries = (float) avgGeometries/(float) numMultiPolygons;
	}
}

void Interior_Point::insert_in_arrangement(CoordinateSequence *cords)
{
	list<Segment_2> segments;

	for (int i = 0; i < (cords->getSize() - 1); i++)
	{
		totalSegments++;
		segments.push_back(Segment_2(Point_2((*cords)[i].x, (*cords)[i].y), Point_2 ((*cords)[i+1].x, (*cords)[i+1].y)));
	}

	if(intersection == true)
		insert(arr, segments.begin(), segments.end());
	else
		insert_non_intersecting_curves(arr, segments.begin(), segments.end());
}

void Interior_Point::read_arrangement()
{
	ifstream input_arrangement(arrangement_input_file);
	if(!input_arrangement.is_open())
	{
		cerr << "interior_point.cpp: Failed to open " << arrangement_input_file << " ..." <<endl;
		exit(1);
	}
	
	IntPoint_Formatter formatter;
	
	input_arrangement >> arr;
	input_arrangement.close();
}

void Interior_Point::write_arrangement()
{
	//write arrangement to a file
	ofstream output(arrangement_output_file);
	output<<arr;
	output.close();	
}

void Interior_Point::write_blocks()
{
	ofstream output(blocks_file);
	IntPoint_Arrangement_2::Face_const_iterator f;

	for(f = arr.faces_begin(); f != arr.faces_end(); ++f)
	{
		// Write the outer boundary of every bounded face.
		if(!f->is_unbounded())
		{
			output<<block_to_wkt(f->outer_ccb())->toString()<<endl;
		}
		// Print the boundary of each of the holes.
		/*Cons_Arrangement_2::Hole_const_iterator hi;
		int index = 1;
		for(hi = f->holes_begin(); hi != f->holes_end(); ++hi, ++index)
		{
			output << "    Hole #" << index << ": ";
			write_ccb (*hi, output);
		}*/
	}
	output.close();	
}

//convert a block in wkt format
Polygon* Interior_Point::block_to_wkt(const IntPoint_Arrangement_2::Ccb_halfedge_const_circulator circ)
{
	CoordinateArraySequence *points = new CoordinateArraySequence;
	IntPoint_Arrangement_2::Ccb_halfedge_const_circulator curr;
	curr = circ;
	Point_2 p = curr->source()->point();
	//to_double is needed if the kernel is exact
	points->add(Coordinate(to_double(p.x()), to_double(p.y())));
	//points->add(Coordinate(p.x(), p.y()));

	do
	{
		IntPoint_Arrangement_2::Halfedge_const_handle he = curr;
		p = he->target()->point();
		//to_double is needed if the kernel is exact
		points->add(Coordinate(to_double(p.x()), to_double(p.y())));
		//points->add(Coordinate(p.x(), p.y()));
	}
	while (++curr != circ);

	PrecisionModel *pm = new PrecisionModel();
	GeometryFactory *fac = new GeometryFactory(pm, srid);

	LinearRing *ring = new LinearRing(points, NULL);
	Polygon *poly = fac->createPolygon(ring, NULL);			//the blocks of the arrangement are only polygons without holes

	return poly;
}

void Interior_Point::arrangement_statistics()
{
	int blockPointsNo = 0, blockNum = 0;
	IntPoint_Arrangement_2::Face_const_iterator f;

	totalBlocks = arr.number_of_faces();
	for(f = arr.faces_begin(); f != arr.faces_end(); ++f)
	{
		if(!f->is_unbounded())
		{
			blockNum++;
			IntPoint_Arrangement_2::Ccb_halfedge_const_circulator circ = f->outer_ccb();
			IntPoint_Arrangement_2::Ccb_halfedge_const_circulator curr = circ;

			blockPointsNo = 1; //count the first point
			do
			{
				blockPointsNo++;
			} while (++curr != circ);
		}
		avgPointsPerBlock += blockPointsNo;
		if(blockNum == 1)
		{
			minPointsPerBlock = blockPointsNo;
			maxPointsPerBlock = blockPointsNo;
		}
		else
		{
			if(blockPointsNo < minPointsPerBlock)
				minPointsPerBlock = blockPointsNo;
			if(blockPointsNo > maxPointsPerBlock)
				maxPointsPerBlock = blockPointsNo;
		}
	}
	avgPointsPerBlock = (float) avgPointsPerBlock / (float) blockNum;
}


/************Computation of the interior/exterior blocks***************/
void Interior_Point::assign_blocks_to_landmarks(int geometries_start, int maxnodeid)
{
	IntPoint_Arrangement_2::Face_const_iterator f;

	//create one interior point for every block	
	for(f = arr.faces_begin(); f != arr.faces_end(); ++f)
	{
		if(!f->is_unbounded())
		{
			blockPoints.push_back(get_interior_point(f));
		}
	}

	interior_exterior_blocks = new char*[maxnodeid + 1];
	for(int i = 0; i <= maxnodeid; i++)
		interior_exterior_blocks[i] = new char[totalBlocks];

	ifstream input_geometries(input_data);
	if(!input_geometries.is_open())
	{
		cerr << "Failed to open " << input_data << " ..." <<endl;
		exit(1);
	}

	input_geometries.seekg(geometries_start, input_geometries.beg);
	
	if(parallel == true)
	{
		find_interior_exterior_blocks_threaded(input_geometries);
	}
	else
	{
		string node, wktGeometry;

		while(true)
		{
			input_geometries >> node;
			if(node[0] == '.')
				break;
			if(node[0] == '\n' || node[0] == '\0')
				continue;
			int node_no = atoi(node.c_str());
			if(debug)
				cout<<"Checking Geometry "<< node <<endl;

			//string interior_blocks, exterior_blocks;
			getline(input_geometries, wktGeometry);
			
			WKTReader *reader = new WKTReader();
			Geometry *geom = reader->read(wktGeometry);	
			if(geom == NULL)
			{
				cerr << "Error reading geometry" << endl;
				exit(1);
			}
			else
			{
				if(geom->getGeometryType() == "Polygon")
				{
					Polygon* poly = dynamic_cast<Polygon*>(geom);
					if(poly == NULL)
					{
						cerr << "Error in dynamic cast" << endl;
						exit(1);
					}
					else
						landmark_interiors(poly, node_no);
				}
				else if(geom->getGeometryType() == "MultiPolygon")
				{
					MultiPolygon* multipoly = dynamic_cast<MultiPolygon*>(geom);
					if(multipoly == NULL)
					{
						cerr << "Error in dynamic cast" << endl;
						exit(1);
					}
					else
						landmark_interiors(multipoly, node_no);
				}
			}
			delete reader;
			delete geom;
		}
	}
	input_geometries.close();
}

void Interior_Point::landmark_interiors(const Polygon *poly, int node)
{
	list<Point_2>::const_iterator it;
	int blocks_counter = 0;

	CoordinateSequence *exteriorCords = poly->getExteriorRing()->getCoordinates();
	
	//iterate over all the blocks of the arrangement
	for(it = blockPoints.begin(); it != blockPoints.end(); it++)
	{
		blocks_counter++;

		if(block_inside(*it, poly, exteriorCords) == true)
		{
			if(debug)
			{
				cout<<"Block: "<<blocks_counter;
				cout<<" Interior block"<<endl;
			}
			interior_exterior_blocks[node][blocks_counter - 1] = INTERIOR;
		}
		else
		{
			if(debug)
			{
				cout<<"Block: "<<blocks_counter;
				cout<<" Exterior block"<<endl;
			}
			interior_exterior_blocks[node][blocks_counter - 1] = EXTERIOR;
		}
	}
	if(debug)
	{
		cout<<"Block: "<<totalBlocks;
		cout<<" Exterior block"<<endl;
	}
	interior_exterior_blocks[node][totalBlocks - 1] = EXTERIOR;
	
	delete exteriorCords;
}

void Interior_Point::landmark_interiors(const MultiPolygon *multipoly, int node)
{
	list<Point_2>::const_iterator it;

	for(int numGeo = 0; numGeo < multipoly->getNumGeometries(); numGeo++)
	{
		const Polygon *poly = dynamic_cast<const Polygon*>(multipoly->getGeometryN(numGeo));
		if(poly == NULL)
		{
			cerr << "Error in dynamic cast" << endl;
			exit(1);
		}

		CoordinateSequence *exteriorCords = poly->getExteriorRing()->getCoordinates();
		int blocks_counter = 0;
		//iterate over all the blocks of the arrangement
		for(it = blockPoints.begin(); it != blockPoints.end(); it++)
		{
			if(numGeo == 0)
				interior_exterior_blocks[node][blocks_counter] = UNKNOWN;
			else if(interior_exterior_blocks[node][blocks_counter] == true)
			{
				blocks_counter++;
				continue;
			}
			if(block_inside(*it, poly, exteriorCords) == true)
			{
				if(debug)
				{
					cout<<"Block: "<<blocks_counter + 1;
					cout<<" Interior block"<<endl;
				}
				interior_exterior_blocks[node][blocks_counter] = INTERIOR;
			}
			else if(numGeo == (multipoly->getNumGeometries() - 1))
			{
				if(interior_exterior_blocks[node][blocks_counter] == UNKNOWN)
				{
					if(debug)
					{
						cout<<"Block: "<<blocks_counter + 1;
						cout<<" Exterior block"<<endl;
					}
					interior_exterior_blocks[node][blocks_counter] = EXTERIOR;
				}
			}
			blocks_counter++;
		}
		delete exteriorCords;
	}
	if(debug)
	{
		cout<<"Block: "<<totalBlocks;
		cout<<" Exterior block"<<endl;
	}
	interior_exterior_blocks[node][totalBlocks - 1] = EXTERIOR;
}

//takes as input a polygon and an interior point of a block and decides whether
//the block lies inside or outside the polygon
bool Interior_Point::block_inside(Point_2 blockPoint, const Polygon *poly, const CoordinateSequence *exteriorCords)
{
	Kernel traits;
	CGAL::Bounded_side side;
	list<Point_2>  points;

	for(int i = 0; i < (exteriorCords->getSize() - 1); i++)
	{
		points.push_back(Point_2((*exteriorCords)[i].x, (*exteriorCords)[i].y));
	}

	//check the point of the block
	side = CGAL::bounded_side_2(points.begin(), points.end(), blockPoint , traits);

	if(side == CGAL::ON_UNBOUNDED_SIDE)
	{
		return false;
	}
	else if(side == CGAL::ON_BOUNDARY)
	{
		return true;
	}
	//if the point lies at the interior, check if it lies inside a hole
	else if(side == CGAL::ON_BOUNDED_SIDE)
	{
		for(int hole = 0; hole < poly->getNumInteriorRing(); hole++)
		{
			points.clear();
			//create the point sequence of the hole
			CoordinateSequence *holeCords = poly->getInteriorRingN(hole)->getCoordinates();
			for(int j = 0; j < (holeCords->getSize() - 1); j++)
			{
				points.push_back(Point_2((*holeCords)[j].x, (*holeCords)[j].y));
			}
			if(insideHole(points, blockPoint) == true)
			{
				return false;
			}
			delete holeCords;
		}
		//if it reaches here it means it does not belong in a hole
		return true;
	}
}

bool Interior_Point::insideHole(const list <Point_2> &holePoints, Point_2 blockPoint)
{
	Kernel traits;

	CGAL::Bounded_side holeSide = CGAL::bounded_side_2(holePoints.begin(), holePoints.end(), blockPoint, traits);

	//the block does not belong in this hole
	if(holeSide == CGAL::ON_UNBOUNDED_SIDE)
		return false;
	else
		return true;
}

//get an interior point of a block
Point_2 Interior_Point::get_interior_point(const IntPoint_Arrangement_2::Face_const_handle &f)
{
	CDT cdt;
	double l1 = 0.2, l2 = 0.3, l3 = 0.5;
	Polygon_2 pgn;
	Kernel traits;
	Polygon_2::iterator it;

	if(!f->is_unbounded())
	{
		IntPoint_Arrangement_2::Ccb_halfedge_const_circulator circ, curr;
		circ = f->outer_ccb();
		curr = circ;

		do
		{
			IntPoint_Arrangement_2::Halfedge_const_handle he = curr;
			pgn.push_back(he->source()->point());
		}
		while (++curr != circ);

		int tries = 0;
		for(int i = 1; i < (pgn.size() - 1); i++)
		{
			if(pgn[i] == pgn[i+1] || pgn[i] == pgn[i-1])
				continue;
			Vector v1 = pgn[i] - pgn[i-1];
			Vector v2 = pgn[i+1] - pgn[i];

			double angle = calculate_angle(v1, v2);

			if(angle < CGAL_PI)
			{
				tries++;
				//try to choose a point close to the vertex and hope that it will be inside
				Point_2 interior_point(l1*pgn[i-1].x() + l2*pgn[i].x() + l3*pgn[i+1].x(), l1*pgn[i-1].y() + l2*pgn[i].y() + l3*pgn[i+1].y());
				CGAL::Bounded_side side = CGAL::bounded_side_2(pgn.vertices_begin(), pgn.vertices_end(), interior_point , traits);
				if(side == CGAL::ON_BOUNDED_SIDE)
				{
					return interior_point;
				}
				if(tries > MAXTRIES)
					break;
			}
		}
		
		do
		{
			IntPoint_Arrangement_2::Halfedge_const_handle he = curr;

			cdt.insert_constraint(CDT_Point(he->source()->point()), CDT_Point(he->target()->point()));
		}
		while (++curr != circ);

		CDT::Face_handle triangle = cdt.finite_faces_begin();
		Point_2 interior_point(l1*triangle->vertex(0)->point().x() + l2*triangle->vertex(1)->point().x() + l3*triangle->vertex(2)->point().x(), l1*triangle->vertex(0)->point().y() + l2*triangle->vertex(1)->point().y() + l3*triangle->vertex(2)->point().y());
		
		return interior_point;
	}
}

/************Parallel computation of interior/exterior blocks**********/
void polygon_interiors_threaded(Interior_Point *method, WKTReader *reader, Geometry *geom, int node, thread_result *res)
{
	list<Point_2>::const_iterator it;
	int blocks_counter = 0;

	res->node = node;
	Polygon* poly = dynamic_cast<Polygon*>(geom);
	if(poly == NULL)
	{
		cerr << "Error in dynamic cast" << endl;
		exit(1);
	}
	else
	{
		CoordinateSequence *exteriorCords = poly->getExteriorRing()->getCoordinates();		
		//iterate over all the blocks of the arrangement
		for(it = shared_blockPoints.begin(); it != shared_blockPoints.end(); it++)
		{
			blocks_counter++;
			if(method->block_inside(*it, poly, exteriorCords) == true)
			{
				res->blocks[blocks_counter - 1] = INTERIOR;
			}	
			else
			{
				res->blocks[blocks_counter - 1] = EXTERIOR;
			}
		}
		res->blocks[totalBlocks - 1] = EXTERIOR;
		delete exteriorCords;	
	}
	delete geom;
	delete reader;
}

void multipolygon_interiors_threaded(Interior_Point *method, WKTReader *reader, Geometry *geom, int node, thread_result *res)
{
	list<Point_2>::const_iterator it;

	res->node = node;
	MultiPolygon* multipoly = dynamic_cast<MultiPolygon*>(geom);
	if(multipoly == NULL)
	{
		cerr << "Error in dynamic cast" << endl;
		exit(1);
	}
	else
	{
		for(int numGeo = 0; numGeo < multipoly->getNumGeometries(); numGeo++)
		{
			const Polygon *poly = dynamic_cast<const Polygon*>(multipoly->getGeometryN(numGeo));
			if(poly == NULL)
			{
				cerr << "Error in dynamic cast" << endl;
				exit(1);
			}

			CoordinateSequence *exteriorCords = poly->getExteriorRing()->getCoordinates();
			int blocks_counter = 0;
			//iterate over all the blocks of the arrangement
			for(it = shared_blockPoints.begin(); it != shared_blockPoints.end(); it++)
			{
				if(numGeo == 0)
					res->blocks[blocks_counter] = UNKNOWN;
				else if(res->blocks[blocks_counter] == UNKNOWN)
				{
					blocks_counter++;
					continue;
				}
				if(method->block_inside(*it, poly, exteriorCords) == true)
					res->blocks[blocks_counter] = INTERIOR;				
				else if(numGeo == (multipoly->getNumGeometries() - 1))
				{
					if(res->blocks[blocks_counter] == UNKNOWN)
						res->blocks[blocks_counter] = EXTERIOR;
				}
				blocks_counter++;
			}
			delete exteriorCords;
		}
		res->blocks[totalBlocks - 1] = EXTERIOR;
	}
	delete geom;
	delete reader;
}

void Interior_Point::find_interior_exterior_blocks_threaded(ifstream &input_geometries)
{
	string node, wktGeometry;
	shared_blockPoints = blockPoints;

	// schedule threads on finding the interior/exterior blocks of a polygon
	while(true)
	{
		input_geometries >> node;
		if(node[0] == '.')
			break;
		if(node[0] == '\n' || node[0] == '\0')
			continue;		
			
		int node_no = atoi(node.c_str());
		//string interior_blocks, exterior_blocks;
		
		getline(input_geometries, wktGeometry);
		
		thread_result *res = new thread_result(); // initialize thread information
		tres.push_back(res);
		
		WKTReader *reader = new WKTReader();
		Geometry *geom = reader->read(wktGeometry);	
		if(geom == NULL)
		{
			cerr << "Error reading geometry" << endl;
			exit(1);
		}
		else
		{
			if(geom->getGeometryType() == "Polygon")
			{	
				workers.push_back(new boost::thread(polygon_interiors_threaded, this, reader, geom, node_no, res));
			}
			else if(geom->getGeometryType() == "MultiPolygon")
			{
				workers.push_back(new boost::thread(multipolygon_interiors_threaded, this, reader, geom, node_no, res));
			}
		}
	}

	if(debug)
	{
		cout << "main thread: waiting for threads to finish...\n";
		cout << "main thread: you'd better have a look at htop\n";
	}

	// wait all threads to finish their work
	wait_threads(workers);
	
	// do something with the results
	for(list<thread_result*>::iterator it = tres.begin(); it != tres.end(); ++it) 
	{
		if(debug)
			cout<<"Checking Geometry "<<(*it)->node<<endl;
		for(int i = 0; i < totalBlocks; i++)
		{
			if(debug)
				cout<<"Block "<<i+1<<" "<<(*it)->blocks[i]<<endl;
			interior_exterior_blocks[(*it)->node][i] = (*it)->blocks[i];		
		}
	}
	// de-allocate memory
	free_threads(workers);
	free_thread_results(tres);
}

bool Interior_Point::check_validity(const IntPoint_Arrangement_2 &arr)
{
	IntPoint_Arrangement_2::Face_const_iterator f;

	for(f = arr.faces_begin(); f != arr.faces_end(); ++f)
	{
		if(!f->is_unbounded())
		{
			Polygon *poly = block_to_wkt(f->outer_ccb());
			if(poly->isValid() == false)
				return false;
		}
	}
	return true;
}

bool Interior_Point::is_valid(const IntPoint_Arrangement_2::Face_const_iterator &f)
{
       Polygon *poly = block_to_wkt(f->outer_ccb());
       if(poly->isValid() == false)
               return false;
       else
               return true;
}

double calculate_angle(Vector v1, Vector v2)
{
	//to_double is needed if the kernel is exact
	double angle = acos(to_double((v1*v2)/ (v1*v1) * (v2*v2)));
	
	//double angle = acos((v1*v2)/ (v1*v1) * (v2*v2));
	
	return angle;
}
