/**
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * Copyright (C) 2013, 2014 Department of Informatics and
 * Telecommunications, National and Kapodistrian University of Athens
 *
 *      Author: Stella Giannakopoulou <sgian@di.uoa.gr>
 */
#include "method.h"

int totalBlocks = 0;

void Method::hybrid_consistency(int geometries_start, int maxnodeid)
{
	ifstream input_geometries(input_data);
	if(!input_geometries.is_open())
	{
		cerr << "split.cpp: Failed to open " << input_data << " ..." <<endl;
		exit(1);
	}
	//seek to the position of the input file where the landmarks start
	input_geometries.seekg(geometries_start, input_geometries.beg);

	struct timeval arr_timer1, arr_timer2;
	gettimeofday(&arr_timer1, NULL);
	double arr_start = arr_timer1.tv_sec + (arr_timer1.tv_usec / (double) 1000000);
	
	create_arrangement(input_geometries);
	gettimeofday(&arr_timer2, NULL);
	double arr_end = arr_timer2.tv_sec + (arr_timer2.tv_usec / (double) 1000000);
	
	if(read_arrangement_from_file == true)
		read_arrangement();
	else
	{
		if(print_time)
			cout << "Construction of the arrangement completed in " << arr_end - arr_start << " seconds." << endl;
		consistency_timer += (arr_end - arr_start);
	}
		
	input_geometries.close();
	
	//print some statistics
	arrangement_statistics();
	if(statistics == true)
		print_statistics();

	if(write_blocks_to_file == true)
		//write resulting blocks to a file
		write_blocks();

	struct timeval ie_timer1, ie_timer2;
	gettimeofday(&ie_timer1, NULL);
	double ie_start = ie_timer1.tv_sec + (ie_timer1.tv_usec / (double) 1000000);

	assign_blocks_to_landmarks(geometries_start, maxnodeid);
	assign_blocks_to_variables(maxnodeid);

	gettimeofday(&ie_timer2, NULL);
	double ie_end = ie_timer2.tv_sec + (ie_timer2.tv_usec / (double) 1000000);

	consistency_timer += (ie_end - ie_start);
	
	if(print_time)
		cout<<"Block assignement completed in "<<ie_end - ie_start<<" seconds"<<endl;
	
	if(write_arrangement_to_file == true)
		write_arrangement();
}

void Method::assign_blocks_to_variables(int maxnodeid)
{
	int node;

	for(int i = 0; i <= maxnodeid; i++)
	{
		if(node_type[i] == 'v')
		{
			initialize_blocks(i);
			for(int j = 0; j <= maxnodeid; j++)
			{
				node = j;
				if(node_type[node] == 'l')
				{
					//if the landmark is contained in the variable, then all its interior blocks
					//are interior blocks of the variable too
					if(renz_matrix[node][i] == (PP) || renz_matrix[i][node] == (PPI))		//PP relation
						set_landmarks_interior_interior(i, node);

					//if the landmark is disconnected or externally connected from the variable, 
					//then all its interior blocks are exterior blocks of the variable
					else if(renz_matrix[node][i] == (DR))				//DR relation
						set_landmarks_exterior_interior(i, node);

					//if the variable is contained in the landmark, then all the exterior blocks
					//of the landmark are exterior blocks of the variable too
					else if(renz_matrix[i][node] == (PP) || renz_matrix[node][i] == (PPI))
						set_landmarks_exterior_exterior(i, node);
				}
			}
		}
	}
}

//initially the position of each block is unknown
void Method::initialize_blocks(int var)
{
	for(int i = 0; i < totalBlocks; i++)
	{
		interior_exterior_blocks[var][i] = UNKNOWN;	
	}
}

/*set the interior blocks of the constant as interior blocks 
 * of the landmark*/
void Method::set_landmarks_interior_interior(int var, int cons)
{
	for(int i = 0; i < totalBlocks; i++)
	{
		if(interior_exterior_blocks[cons][i] == INTERIOR)
			interior_exterior_blocks[var][i] = INTERIOR;
	}
}

/*set the interior blocks of the landmark as exterior blocks
 * of the variable*/
void Method::set_landmarks_exterior_interior(int var, int cons)
{
	for(int i = 0; i < totalBlocks; i++)
	{
		if(interior_exterior_blocks[cons][i] == INTERIOR)
			interior_exterior_blocks[var][i] = EXTERIOR;
	}
}

/*set the exterior blocks of the landmark as exterior blocks
 * of the variable*/
void Method::set_landmarks_exterior_exterior(int var, int cons)
{
	for(int i = 0; i < totalBlocks; i++)
	{
		if(interior_exterior_blocks[cons][i] == EXTERIOR)
			interior_exterior_blocks[var][i] = EXTERIOR;
	}
}

void Method::print_statistics()
{
	cout<<"Max number of points per polygon: "<<maxPolygonPoints<<endl;
	cout<<"Min number of points per polygon: "<<minPolygonPoints<<endl;
	cout<<"Average number of points per polygon: "<<avgPolygonPoints<<endl;

	cout<<"Max number of points per multipolygon: "<<maxMultiPolygonPoints<<endl;
	cout<<"Min number of points per multipolygon: "<<minMultiPolygonPoints<<endl;
	cout<<"Average number of points per multipolygon: "<<avgMultiPolygonPoints<<endl;

	cout<<"Max number of geometries per multipolygon: "<<maxGeometries<<endl;
	cout<<"Min number of geometries per multipolygon: "<<minGeometries<<endl;
	cout<<"Average number of geometries per multipolygon: "<<avgGeometries<<endl;
	
	cout<<"Total inserted segments: "<<totalSegments<<endl;
	cout<<"Total blocks: "<<totalBlocks<<endl;

	cout<<"Max points per block: "<<maxPointsPerBlock<<endl;
	cout<<"Min points per block: "<<minPointsPerBlock<<endl;
	cout<<"Average points per block: "<<avgPointsPerBlock<<endl;
}

