/**
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * Copyright (C) 2013, 2014 Department of Informatics and
 * Telecommunications, National and Kapodistrian University of Athens
 *
 *      Author: Stella Giannakopoulou <sgian@di.uoa.gr>
 */
 
#ifndef METHOD_H
#define METHOD_H

#define TRUE "1"
#define FALSE "0"
#define INTERIOR 'i'
#define EXTERIOR 'e'
#define UNKNOWN 'u'
#define HOLE '1'
#define NOT_HOLE '0'

#include "utils.h"
#include <boost/thread/thread.hpp>

/**Options**/
extern RELTYPE **renz_matrix;
extern const char *input_data;
extern bool intersection, statistics, debug, print_time, parallel;
extern const char *blocks_file, *arrangement_output_file, *arrangement_input_file;
extern bool write_blocks_to_file, write_arrangement_to_file, read_arrangement_from_file;
extern int srid;
extern int totalBlocks, totalGeometries;
extern char *node_type;
extern char **interior_exterior_blocks;
extern double consistency_timer;

typedef struct thread_result 
{
	char *blocks;
	int node;

	thread_result() 
	{
		blocks = new char[totalBlocks];
		node = -1;
	}
} thread_result;

class Method
{
	public:
		string method;
		list<boost::thread*> workers;	/**< list of worker threads */
		list<thread_result*> tres;
		
		//statistics
		int maxPolygonPoints, minPolygonPoints, maxMultiPolygonPoints, minMultiPolygonPoints, maxGeometries, minGeometries, minPointsPerBlock, maxPointsPerBlock, totalSegments, totalGeometries;
		float avgPolygonPoints, avgMultiPolygonPoints, avgGeometries, avgPointsPerBlock;
			
		Method()
		{
			//initialize statistics
			totalSegments = 0;
			totalGeometries = 0;
			avgPolygonPoints = 0;
			avgMultiPolygonPoints = 0; 
			avgGeometries = 0; 
			maxGeometries = 0;
			minGeometries = 0;
			avgPointsPerBlock = 0;	
			maxPolygonPoints = 0;
			minPolygonPoints = 0;
			maxMultiPolygonPoints = 0;
			minMultiPolygonPoints = 0;
			avgPolygonPoints = 0;
			avgMultiPolygonPoints = 0;
		}
		void hybrid_consistency(int geometries_start, int maxnodeid);
		void print_statistics();
		virtual void create_arrangement(ifstream &input_geometries) = 0;
		virtual void read_arrangement() = 0;
		virtual void write_arrangement() = 0;
		virtual void write_blocks() = 0;
		virtual void arrangement_statistics() = 0;
		virtual void assign_blocks_to_landmarks(int geometries_start, int maxnodeid) = 0;
		void assign_blocks_to_variables(int maxnodeid);
		void initialize_blocks(int var);
		void set_landmarks_interior_interior(int var, int cons);
		void set_landmarks_exterior_interior(int var, int cons);
		void set_landmarks_exterior_exterior(int var, int cons);

		
		inline void free_threads(std::list<boost::thread*> &workers)
		{
			for (std::list<boost::thread*>::iterator it = workers.begin(); it != workers.end(); ++it)
			{
				delete *it;
			}
			workers.clear();
		}

		inline void free_thread_results(std::list<thread_result*> &results) 
		{
			for (std::list<thread_result*>::iterator it = results.begin(); it != results.end(); ++it)
			{
				delete [] (*it)->blocks;
				delete *it;
			}
			results.clear();
		}

		inline void wait_threads(std::list<boost::thread*> &workers) 
		{
			for (std::list<boost::thread*>::iterator it = workers.begin(); it != workers.end(); ++it)
			{
				(*it)->join();

			} // all threads stopped
		}
};

/*****Conditions**********/
bool check_theorem(int maxnodeid);
bool variable_exterior_condition(int var);
bool partial_overlap_conditions(int node1, int node2);
bool proper_part_conditions(int node1, int node2);
#if defined(RCC5)
bool disjoint_condition(int node1, int node2);
#endif
int get_blocks_size(int node, char type);
int get_union_blocks_size(int node1, int node2, char type1, char type2);
vector<int> get_blocks(int node, char type);

#endif /*METHOD_H*/
