/**
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * Copyright (C) 2013, 2014 Department of Informatics and
 * Telecommunications, National and Kapodistrian University of Athens
 *
 *      Author: Stella Giannakopoulou <sgian@di.uoa.gr>
 */

#include "overlay.h"

void Overlay::create_arrangement(ifstream &input_geometries)
{
	int numGeos = 0, numPolygons = 0, numMultiPolygons = 0, numPolygonPoints, numMultiPolygonPoints;
	string node;
	string wktGeometry;													//the string WKT literal of the current geometry read

	if(debug)
		cout<<"Creating arrangement..."<<endl;
	int i = 0;
	while(true)
	{
		input_geometries >> node;
		if(node[0] == '.')
			break;
		if(node[0] == '\n' || node[0] == '\0')
			continue;
		getline(input_geometries, wktGeometry);

		node_type[atoi(node.c_str())] = 'l';                     		//the node is a landmark

		WKTReader *reader = new WKTReader();
		Geometry *geom = reader->read(wktGeometry);

		if(!read_arrangement_from_file)
		{
			if(geom == NULL)
			{
				cerr << "overlay.cpp: Error reading geometry" << endl;
				exit(1);
			}
			else
			{
				if(debug)
					cout << "Geometry: "<< i+1 <<endl;
				if(geom->getGeometryType() == "Polygon")
				{
					numPolygons++;
					Polygon* poly = dynamic_cast<Polygon*>(geom);				

					if(poly == NULL)
					{
						cerr << "overlay.cpp: Error in dynamic cast" << endl;
						exit(1);
					}
					numPolygonPoints = poly->getNumPoints();
					avgPolygonPoints += numPolygonPoints;

					if(numPolygons == 1)
					{
						minPolygonPoints = numPolygonPoints;
						maxPolygonPoints = numPolygonPoints;
					}
					else
					{
						if(numPolygonPoints < minPolygonPoints)
							minPolygonPoints = numPolygonPoints;

						if(numPolygonPoints > maxPolygonPoints)
							maxPolygonPoints = numPolygonPoints;
					}

					CoordinateSequence *cords = poly->getExteriorRing()->getCoordinates();

					insert_in_arrangement(arr, cords);
					delete cords;

					for(int hole = 0; hole < poly->getNumInteriorRing(); hole++)
					{
						CoordinateSequence *cords = poly->getInteriorRingN(hole)->getCoordinates();

						insert_in_arrangement(arr, cords);
						delete cords;
					}
				}
				else if(geom->getGeometryType() == "MultiPolygon")
				{
					numMultiPolygons++;
					//cout<<"Type: "<<geom->getGeometryType()<<" and contains "<<geom->getNumGeometries() <<" geometries"<<endl;
					MultiPolygon* multipoly = dynamic_cast<MultiPolygon*>(geom);
					if(multipoly == NULL)
					{
						cerr << "overlay.cpp: Error in dynamic cast" << endl;
						exit(1);
					}
					numGeos = multipoly->getNumGeometries();
					avgGeometries+=numGeos;
					numMultiPolygonPoints = 0;
					for(int numGeo = 0; numGeo < multipoly->getNumGeometries(); numGeo++)
					{
						const Polygon *poly = dynamic_cast<const Polygon*>(multipoly->getGeometryN(numGeo));
						if(poly == NULL)
						{
							cerr << "overlay.cpp: Error in dynamic cast" << endl;
							exit(1);
						}
						numMultiPolygonPoints += poly->getNumPoints();
						
						CoordinateSequence *cords = poly->getExteriorRing()->getCoordinates();
							
						insert_in_arrangement(arr, cords);
						delete cords;
						
						for(int hole = 0; hole < poly->getNumInteriorRing(); hole++)
						{
							CoordinateSequence *cords = poly->getInteriorRingN(hole)->getCoordinates();
							
							insert_in_arrangement(arr, cords);
							delete cords;
						}
					}
					avgMultiPolygonPoints+=numMultiPolygonPoints;
					if(numMultiPolygons == 1)
					{
						minMultiPolygonPoints = numMultiPolygonPoints;
						maxMultiPolygonPoints = numMultiPolygonPoints;

						minGeometries = numGeos;
						maxGeometries = numGeos;
					}
					else
					{
						if(numMultiPolygonPoints < minMultiPolygonPoints)
							minMultiPolygonPoints = numMultiPolygonPoints;

						if(numMultiPolygonPoints > maxMultiPolygonPoints)
							maxMultiPolygonPoints = numMultiPolygonPoints;

						if(numGeos < minGeometries)
							minGeometries = numGeos;
						if(numGeos > maxGeometries)
							maxGeometries = numGeos;
					}
				}
			}
		}	
		i++;
		delete reader;
		delete geom;
	}
	if(numPolygons > 0)
		avgPolygonPoints = (float) avgPolygonPoints/(float) numPolygons;
	if(numMultiPolygons > 0)
	{
		avgMultiPolygonPoints = (float) avgMultiPolygonPoints/(float) numMultiPolygons;
		avgGeometries = (float) avgGeometries/(float) numMultiPolygons;
	}
}

void Overlay::insert_in_arrangement(Ov_Arrangement_2 &arrang, CoordinateSequence *cords)
{
	list<Segment_2> segments;

	for (int i = 0; i < (cords->getSize() - 1); i++)
	{
		totalSegments++;
		segments.push_back(Segment_2(Point_2((*cords)[i].x, (*cords)[i].y), Point_2 ((*cords)[i+1].x, (*cords)[i+1].y)));
	}

	if(intersection == true)
		insert(arrang, segments.begin(), segments.end());
	else
		insert_non_intersecting_curves(arrang, segments.begin(), segments.end());
}

void Overlay::arrangement_statistics()
{
	int blockPoints = 0, blockNum = 0;
	Ov_Arrangement_2::Face_const_iterator f;

	totalBlocks = arr.number_of_faces();
	for(f = arr.faces_begin(); f != arr.faces_end(); ++f)
	{
		if(!f->is_unbounded())
		{
			blockNum++;
			Ov_Arrangement_2::Ccb_halfedge_const_circulator circ = f->outer_ccb();
			Ov_Arrangement_2::Ccb_halfedge_const_circulator curr = circ;

			blockPoints = 1; //count the first point
			do
			{
				blockPoints++;
			} while (++curr != circ);
		}
		avgPointsPerBlock += blockPoints;
		if(blockNum == 1)
		{
			minPointsPerBlock = blockPoints;
			maxPointsPerBlock = blockPoints;
		}
		else
		{
			if(blockPoints < minPointsPerBlock)
				minPointsPerBlock = blockPoints;
			if(blockPoints > maxPointsPerBlock)
				maxPointsPerBlock = blockPoints;
		}
	}
	avgPointsPerBlock = (float) avgPointsPerBlock / (float) blockNum;
}

void Overlay::read_arrangement()
{	
	ifstream input_arrangement(arrangement_input_file);
	if(!input_arrangement.is_open())
	{
		cerr << "overlay.cpp: Failed to open " << arrangement_input_file << " ..." <<endl;
		exit(1);
	}
	//read(arr, input_arrangement, formatter);
	input_arrangement.close();
}

void Overlay::write_arrangement()
{
	//write arrangement to a file	
	ofstream output(arrangement_output_file);
	//write(arr, output, formatter);
	output.close();
}

void Overlay::write_blocks()
{
	ofstream output(blocks_file);
	Ov_Arrangement_2::Face_const_iterator f;

	for(f = arr.faces_begin(); f != arr.faces_end(); ++f)
	{
		// Write the outer boundary of every bounded face.
		if(!f->is_unbounded())
		{
			output<<block_to_wkt(f->outer_ccb())->toString()<<endl;
		}
		// Print the boundary of each of the holes.
		//Ov_Arrangement_2::Hole_const_iterator hi;
		//int index = 1;
		//for(hi = f->holes_begin(); hi != f->holes_end(); ++hi, ++index)
		//{
			//output << "    Hole #" << index << ": ";
			//write_ccb (*hi, output);
		//}
	}
	output.close();
}

//convert a block in wkt format
Polygon* Overlay::block_to_wkt(const Ov_Arrangement_2::Ccb_halfedge_const_circulator circ)
{
	CoordinateArraySequence *points = new CoordinateArraySequence;
	Ov_Arrangement_2::Ccb_halfedge_const_circulator curr;
	curr = circ;
	Point_2 p = curr->source()->point();
	//to_double is needed if the kernel is exact
	points->add(Coordinate(to_double(p.x()), to_double(p.y())));
	//points->add(Coordinate(p.x(), p.y()));

	do
	{
		Ov_Arrangement_2::Halfedge_const_handle he = curr;
		p = he->target()->point();
		//to_double is needed if the kernel is exact
		points->add(Coordinate(to_double(p.x()), to_double(p.y())));
		//points->add(Coordinate(p.x(), p.y()));
	}
	while (++curr != circ);

	PrecisionModel *pm = new PrecisionModel();
	GeometryFactory *fac = new GeometryFactory(pm, srid);

	LinearRing *ring = new LinearRing(points, NULL);
	Polygon *poly = fac->createPolygon(ring, NULL);			//the blocks of the arrangement are only polygons without holes

	return poly;
}

bool Overlay::check_validity(const Ov_Arrangement_2 &arr)
{
	Ov_Arrangement_2::Face_const_iterator f;

	for(f = arr.faces_begin(); f != arr.faces_end(); ++f)
	{
		if(!f->is_unbounded())
		{
			Polygon *poly = block_to_wkt(f->outer_ccb());
			if(poly->isValid() == false)
				return false;
		}
	}
	return true;
}

bool Overlay::is_valid(const Ov_Arrangement_2::Face_const_iterator &f)
{
       Polygon *poly = block_to_wkt(f->outer_ccb());
       if(poly->isValid() == false)
               return false;
       else
               return true;
}

/************Computation of the interior/exterior blocks***************/
void Overlay::assign_blocks_to_landmarks(int geometries_start, int maxnodeid)
{
	Ov_Arrangement_2::Face_iterator f;
	string wktGeometry, node;

	interior_exterior_blocks = new char*[maxnodeid + 1];
	for(int i = 0; i <= maxnodeid; i++)
		interior_exterior_blocks[i] = new char[totalBlocks];

	ifstream input_geometries(input_data);
	if(!input_geometries.is_open())
	{
		cerr << "overlay.cpp: Failed to open " << input_data << " ..." <<endl;
		exit(1);
	}

	if(debug)
		cout<<"Assigning blocks to polygons: "<<endl;
	input_geometries.seekg(geometries_start, input_geometries.beg);

	//set the appropriate data needed for the overlay
	//set 1 + block number for every bounded block
	//and 0 + block number for the unbounded block
	int blocks_counter = 0;
	for(f = arr.faces_begin(); f != arr.faces_end(); ++f)
	{
		ostringstream ostr;
		if(!f->is_unbounded())
		{
			blocks_counter++;
			ostr<<blocks_counter;
			f->set_data("1" + ostr.str());
		}
		else
		{
			ostr<<totalBlocks;
			f->set_data("0" + ostr.str());
		}
	}
	int i = 0;
	while(true)
	{
		input_geometries >> node;
		int node_no = atoi(node.c_str());
		if(node[0] == '.')
			break;
		if(node[0] == '\n' || node[0] == '\0')
			continue;

		getline(input_geometries, wktGeometry);
		WKTReader *reader = new WKTReader();						// the reader for WKT geometries 
		Geometry *geom = reader->read(wktGeometry);

		if(geom == NULL)
		{
			cerr << "overlay.cpp:  line " << i + 1 << ": error reading geometry" << endl;
			exit(1);
		}
		else
		{
			if(debug)
				cout<<"Checking Geometry "<< node <<endl;

			Ov_Arrangement_2 arr_single_geo;
			if(geom->getGeometryType() == "Polygon")
			{
				Polygon* poly = dynamic_cast<Polygon*>(geom);
				if(poly == NULL)
				{
					cerr << "overlay.cpp: Error in dynamic cast" << endl;
					exit(1);
				}
				else
				{
					//insert first the holes in order to mark the respective
					//blocks with -1 --exterior because they belong to a hole
					for(int hole = 0; hole < poly->getNumInteriorRing(); hole++)
					{
						CoordinateSequence *cords = poly->getInteriorRingN(hole)->getCoordinates();
						
						insert_in_arrangement(arr_single_geo, cords);
						delete cords;
					}
					
					//mark all the holes of the polygon with -1
					for(f = arr_single_geo.faces_begin(); f != arr_single_geo.faces_end(); ++f)
					{
						if(!f->is_unbounded())
							f->set_data("-1");
					}
					
					//then insert the outer boundary of the polygon
					CoordinateSequence *cords = poly->getExteriorRing()->getCoordinates();
					insert_in_arrangement(arr_single_geo, cords);
					delete cords;
				}
			}
			else if(geom->getGeometryType() == "MultiPolygon")
			{
				MultiPolygon* multipoly = dynamic_cast<MultiPolygon*>(geom);
				if(multipoly == NULL)
				{
					cerr << "overlay.cpp: Error in dynamic cast" << endl;
					exit(1);
				}
				else
				{
					//insert first the holes in order to mark the respective
					//blocks with -1 --exterior because they belong to a hole
					for(int numGeo = 0; numGeo < multipoly->getNumGeometries(); numGeo++)
					{
						const Polygon *poly = dynamic_cast<const Polygon*>(multipoly->getGeometryN(numGeo));
						if(poly == NULL)
						{
							cerr << "overlay.cpp: Error in dynamic cast" << endl;
							exit(1);
						}
						for(int hole = 0; hole < poly->getNumInteriorRing(); hole++)
						{
							CoordinateSequence *cords = poly->getInteriorRingN(hole)->getCoordinates();
							
							insert_in_arrangement(arr_single_geo, cords);
							delete cords;
						}
					}

					//mark all the holes of the polygon with -1
					for(f = arr_single_geo.faces_begin(); f != arr_single_geo.faces_end(); ++f)
					{
						if(!f->is_unbounded())
							f->set_data("-1");
					}

					//then proceed with the outer boundary
					for(int numGeo = 0; numGeo < multipoly->getNumGeometries(); numGeo++)
					{
						const Polygon *poly = dynamic_cast<const Polygon*>(multipoly->getGeometryN(numGeo));
						if(poly == NULL)
						{
							cerr << "overlay.cpp: Error in dynamic cast" << endl;
							exit(1);
						}
						CoordinateSequence *cords = poly->getExteriorRing()->getCoordinates();
						insert_in_arrangement(arr_single_geo, cords);
						delete cords;
					}
				}
			}
			//set the appropriate data needed for the overlay
			//1 for every bounded block and 0 for the unbounded block
			for(f = arr_single_geo.faces_begin(); f != arr_single_geo.faces_end(); ++f)
			{
				if(!f->is_unbounded() && f->data() != "-1")
					f->set_data("1");
				else
					f->set_data("0");
			}
			// Compute the overlay of the two arrangements
			Ov_Arrangement_2 overlay_arr;
			Overlay_traits overlay_traits;

			overlay(arr, arr_single_geo, overlay_arr, overlay_traits);

			// Go over the faces of the overlaid arrangement and their labels.
			Ov_Arrangement_2::Face_iterator  res_fit;

			for(res_fit = overlay_arr.faces_begin(); res_fit != overlay_arr.faces_end(); ++res_fit)
			{
				string data = res_fit->data();
				int block_num = atoi(data.substr(1).c_str());
				//cout << data.substr(1) << " ";
				if(data.substr(0, 1) == "2")
				{
					if(debug)
						cout<<" Interior block "<<block_num<<endl;
					interior_exterior_blocks[node_no][block_num - 1] = INTERIOR;
				}
				else
				{
					if(debug)
						cout<<" Exterior block "<<block_num<<endl;
					interior_exterior_blocks[node_no][block_num - 1] = EXTERIOR;
				}
			}
		}
		i++;
		delete reader;
		delete geom;
	}
	input_geometries.close();
}

