/**
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * Copyright (C) 2013, 2014 Department of Informatics and
 * Telecommunications, National and Kapodistrian University of Athens
 *
 *      Author: Stella Giannakopoulou <sgian@di.uoa.gr>
 */

#include "qq_rcc_consistency.h"

int main(int argc, char *argv[])
{
	const string exe = argv[0];											/* the name of this executable */
	string report_bugs = "Author: Stella Giannakopoulou <sgian@di.uoa.gr>\nBugs: Please report any bugs to <sgian@di.uoa.gr>\n";
	string copyright="Copyright (C) 2013, 2014 Department of Informatics and Telecommunications,\nNational and Kapodistrian University of Athens.\n";
	copyright+="This program is free software. You can redistribute it and/or modify it\nunder the terms of the GNU General Public License.";

	exe_dir = exe.substr(0, exe.find_last_of("/\\"));
	input_data = argv[argc - 1];
	srid = 4326;
	blocks_method="consolidated";
	debug = false;
	
	namespace po = boost::program_options; 
	po::options_description desc("OPTIONS");
	desc.add_options()
    ("help,h", "produce help message")
    ("input,i", po::value<string>(), "insert option of a segment in the arrangement \"intersectingCurves\" or \"disjointCurves\" (default value is intersectingCurves)")
    ("heuristic,e", po::value<string>(), "\"lcr\" and/or \"sd\": heuristic for least constraining relation and relation with the smallest decomposition")
    ("method,m", po::value<string>(), "\"consolidated\", \"interior_point\" or \"overlay\" (default consolidated)")
    ("srid,s", po::value<int>(), "the srid of the given geometries (default 4326)")
    ("read,r", po::value<string>(), "read the arrangement from an input file and don't construct it")
    ("write,w", po::value<string>(), "output the arrangement in a file for later use")
    ("output,o", po::value<string>(), "output the blocks of the arrangement in a file in WKT format")
    ("parallel,p", "parallelize the procedure of finding the interior/exterior blocks")
    ("statistics,c", "print information about the input polygons and the resulting arrangement")
    ("debug,d", "print debug information")
    ("time,t", "print time information");

	po::variables_map vm;
	int argsNo = 0;
	
	try
	{
		po::store(po::parse_command_line(argc, argv, desc), vm);
		po::notify(vm);
		
		if(argc < 2 || argc > 21)
		{	
			cerr << "Usage:\n"<<endl;
			cerr<< exe << " [OPTIONS] FILE\n" << endl;
			cerr<<desc<<endl;
			cerr<<report_bugs<<endl;
			cerr<<copyright<<endl;
			return 1;
		}
		if(vm.count("help"))
		{
			cerr << "Usage:\n"<<endl;
			cerr<< exe << " [OPTIONS] FILE\n" << endl;
			cerr<<desc<<endl;
			cerr<<report_bugs<<endl;
			cerr<<copyright<<endl;
			return 1;
		}
		if(vm.count("input"))
		{
			argsNo+=2;
			string value = vm["input"].as<string>();
			if(value == "intersectingCurves")
				intersection = true;
			else if(value == "disjointCurves")
				intersection = false;	
			else
				throw po::error(po::unknown_option(value));
		}
		if(vm.count("heuristic"))
		{
			argsNo++;
			string value = vm["heuristic"].as<string>();
			if(value == "lcr")
			{
				argsNo++;
				lcr = true;
			}	
			if(value == "sd")
			{
				argsNo++;
				sd = false;
			}	
			if((value != "lcr") && (value != "sd"))
				throw po::error(po::unknown_option(value));
		}
		if(vm.count("srid"))
		{
			argsNo+=2;
			srid = vm["srid"].as<int>();
		}	
		if(vm.count("method"))
		{
			argsNo+=2;
			blocks_method = vm["method"].as<string>();
			if(blocks_method != "consolidated" && blocks_method != "interior_point" && blocks_method != "overlay")
				throw po::error(po::unknown_option(blocks_method));
		}
		if(vm.count("read"))
		{
			argsNo+=2;
			arrangement_input_file = (vm["read"].as<string>()).c_str();
			read_arrangement_from_file = true;
		}
		if(vm.count("write"))
		{
			argsNo+=2;
			arrangement_input_file = (vm["write"].as<string>()).c_str();
			write_arrangement_to_file = true;
		}
		if(vm.count("output"))
		{
			argsNo+=2;
			blocks_file = (vm["output"].as<string>()).c_str();
			write_blocks_to_file = true;
		}
		if(vm.count("parallel"))
		{
			argsNo++;
			parallel = true;
		}
		if(vm.count("statistics"))
		{
			argsNo++;
			statistics = true;
		}
		if(vm.count("debug"))
		{
			argsNo++;
			debug = true;
		}
		if(vm.count("time"))
		{
			argsNo++;
			print_time = true;
		}
		
		if(argc != (argsNo + 2))
			throw po::error(po::required_option("FILE"));
	} catch (exception& e) 
	{
		cerr<<"Error: "<<e.what()<< endl;
		return 1;
	}	
	
	if(blocks_method == "overlay" && parallel == true)
	{
		cerr<< "Overlay method is not a thread safe operation ..."<< endl;
		return 1;	
	}	
	check_consistency();
}

/*checks the consisency of a general RC-5 or RCC-8 network that
 * contains constants*/
void check_consistency()
{
	int maxnodeid, inp;
	char csptype[100];
	FILE *infile, *splitfile;
	
	if ((infile = fopen(input_data, "r")) == NULL)
	{
		cerr<<"Cannot open "<<input_data<<" for reading"<<endl;
		exit(-1);
	}

	parsecsp_vsd(infile, MAXCSP-1, 0, &maxnodeid, csptype, &renz_matrix);
	int geometries_start = ftell(infile);
	fclose(infile);
	
	mark = (char **)malloc((maxnodeid + 1)*sizeof(char *));
	if((mark[0] = (char *) malloc((maxnodeid + 1)*(maxnodeid + 1)*sizeof(char))) == NULL)
	{
		cerr<<"No more memory available"<<endl;
		exit(-1);
	}
	for(int i=1; i < (maxnodeid + 1); i++)
		mark[i] = mark[0] + i * (maxnodeid + 1);

	#if defined(RCC5)
	if((splitfile = fopen((exe_dir+"/src/renz-solver/basesplit_rcc5").c_str(), "r")) == NULL)
	#elif defined(RCC8)
	if((splitfile = fopen((exe_dir.c_str()+"/src/renz-solver/basesplit_rcc8").c_str(), "r")) == NULL)
	#endif
	{
		cerr<< "Split file 'baseplit' not found"<<endl;;
		exit(-1);
	}

	for(int i = 0; i < MAXSET; i++)
    {
		for(int j = 0; j < (MAXREL + 1); j++)
		{
			if (fscanf(splitfile, "%d", &inp) != 1)
			{
				fprintf(stderr, "Format error in relation-cover file 'basesplit'\n");
				exit(-1);
			}
			else
				splitinfo[i][j] = inp;
		}
	}

	bool consistency_status;
	if(blocks_method == "consolidated")
	{
		Consolidated_Data method;
		consistency_status = basic_consistency(renz_matrix, method, maxnodeid, infile, csptype, geometries_start);
	}
	else if(blocks_method == "interior_point")
	{
		Interior_Point method;
		consistency_status = basic_consistency(renz_matrix, method, maxnodeid, infile, csptype, geometries_start);
	}
	else if(blocks_method == "overlay")
	{
		Overlay method;
		consistency_status = basic_consistency(renz_matrix, method, maxnodeid, infile, csptype, geometries_start);	
	}
	
	if(consistency_status)
	{
		cout<<"Consistent Network"<<endl;
	}
	else
		cout<<"Inconsistent Network"<<endl;

	cout<<"Total consistency time is: "<<consistency_timer<<" seconds."<<endl;
	free((void *) mark[0]);
	free((void *) mark);
	free((void *) renz_matrix[0]);
	free((void *) renz_matrix);
	
	delete [] node_type;
	if(consistency_check > 0)
	{
		for(int i = 0; i <= maxnodeid; i++) 
		{
			delete [] interior_exterior_blocks[i];
		}
		delete [] interior_exterior_blocks;	
	}
}

/*create recursively possible basic networks and return true if
 * at least one is consistent and false otherwise*/
bool basic_consistency(RELTYPE **csp, Method &method, int maxnodeid, FILE *infile, char *csptype, int geometries_start)
{
	RELTYPE **newcsp;
	int cspsize = maxnodeid+1;
	#if defined(RCC5)
	RELTYPE split[5];
	#elif defined(RCC8)
	RELTYPE split[8];
	#endif
	
	swdebug = 0; // no debug
	pcops = 0;   // path consistency operations
	pcits = 0;   // path consistency iterations

	//check the path consistency of the RCC-5 network, using renz's PC
	struct timeval pc_timer1, pc_timer2;
	gettimeofday(&pc_timer1, NULL);
	double pc_start = pc_timer1.tv_sec + (pc_timer1.tv_usec / (double) 1000000);
	
	initfop();
	int path_consistent = path_cons1(csp, maxnodeid, -1, -1);
	
	gettimeofday(&pc_timer2, NULL);
	double pc_end = pc_timer2.tv_sec + (pc_timer2.tv_usec / (double) 1000000);

	if(!path_consistent)
		return false;
	if(print_time)	
		cout << "Path consistency checking completed in " << pc_end - pc_start << " seconds." << endl;
	consistency_timer += (pc_end - pc_start);
	if(debug)
		printcsp(stdout, maxnodeid, csptype, csp);
	
	newcsp = (RELTYPE **)malloc(cspsize*sizeof(RELTYPE *));
	newcsp[0] = (RELTYPE *)malloc(cspsize*cspsize*sizeof(RELTYPE));
	for(int i = 1; i < cspsize; i++)
		newcsp[i] = newcsp[0] + i*cspsize;
	for(int i = 0; i <= maxnodeid; i++)
	{
		if(backtrack_depth == 0)
		{
			for(int j = 0; j <= maxnodeid; j++)
			{
				if(i <= j)
					newcsp[i][j] = csp[i][j];
				else
					newcsp[i][j] = 1;
			}
		}
		else
		{
			for(int j = 0; j <= maxnodeid; j++)
				newcsp[i][j] = csp[i][j];
		}
	}
	
	int splitnode1, splitnode2;
	
	backtrack_depth++;
	if(split_relation(newcsp, maxnodeid, &splitnode1, &splitnode2) == false)
	{
		gen_split(newcsp[splitnode1][splitnode2], split);
		for(int s = 0; s < MAXREL; s++)
		{
			if(split[s] != 0)
			{
				newcsp[splitnode1][splitnode2] = split[s];
				if(basic_consistency(newcsp, method, maxnodeid, infile, csptype, geometries_start))
				{
					free((void *) newcsp[0]);
					free((void *) newcsp);
					return true;
				}
			}
		}
	}
	else
	{
		free((void *) newcsp[0]);
		free((void *) newcsp);

		for(int i = 0; i <= maxnodeid; i++)
		{
			for(int j = 0; j <= maxnodeid; j++)
				renz_matrix[i][j] = csp[i][j];
		}

		if(consistency_check == 0)
		{
			node_type = new char[maxnodeid + 1];
			for(int i = 0; i <= maxnodeid; i++)
			{
				node_type[i] = 'v';
			}

			//split the plane into non-intersecting blocks
			method.hybrid_consistency(geometries_start, maxnodeid);
			consistency_check++;
		}
		else
		{
			method.assign_blocks_to_variables(maxnodeid);
		}
		struct timeval conditions_timer1, conditions_timer2;
		gettimeofday(&conditions_timer1, NULL);
		double conditions_start = conditions_timer1.tv_sec + (conditions_timer1.tv_usec / (double) 1000000);
		
		//check the conditions of the theorem
		if(check_theorem(maxnodeid) == true)
		{
			gettimeofday(&conditions_timer2, NULL);
			double conditions_end = conditions_timer2.tv_sec + (conditions_timer2.tv_usec / (double) 1000000);
			if(print_time)
				cout << "Conditions' checking completed in " << (conditions_end - conditions_start) << " seconds." << endl;
			consistency_timer += (conditions_end - conditions_start);
			return true;
		}
		else
		{
			gettimeofday(&conditions_timer2, NULL);
			double conditions_end = conditions_timer2.tv_sec + (conditions_timer2.tv_usec / (double) 1000000);
			if(print_time)
			{
				cout<< "Inconsistent network"<<endl;	
				cout << "Conditions' checking completed in " << (conditions_end - conditions_start) << " seconds." << endl;
			}	
			consistency_timer += (conditions_end - conditions_start);
			return false;
		}
	}
	free((void *) newcsp[0]);
	free((void *) newcsp);
}

/*computes the relation to be split
 * returns true if no other relation can be split*/
bool split_relation(RELTYPE **csp, int maxnodeid, int *node1, int *node2)
{
	int defaultval = DEFAULTVAL;
	int temp, minval;
	RELTYPE r;

	minval = defaultval;

	for(int i = 0; i <= maxnodeid; i++)
	{
		for(int j = i+1; j <= maxnodeid; j++)
		{
			if(csp[j][i])
			{
				r = csp[i][j];
				if(splitinfo[r][0] == 1)
				{
					temp=0;
					csp[j][i] = 0;
				}
				else
				{
					if(sd == true)
					{
						temp = splitinfo[r][0];
						if(temp < minval)
						{
							minval = temp;
							*node1 = i;
							*node2 = j;
						}
					}
					else
					{
						*node1 = i;
						*node2 = j;
						return false;
					}
				}
			}
		}
	}
	if(minval == DEFAULTVAL)
		return true;
	else
		return false;
}

/* generate a "split" for RCC-5/RCC-8 relations*/
void gen_split(RELTYPE r, RELTYPE *split)
{
	if(lcr == true)
	{
		for(int i = 1; i <= MAXREL; i++)
			split[i-1] = splitinfo[r][i];
	}
	else
	{
		int i;
	
		for(i = 0; i < MAXREL; i++)
			split[i] = 0;
		i = 0;
		#if defined(RCC5)
		if(r & PO)
			split[i++] = PO;
		if(r & DR)
			split[i++] = DR;
		if(r & PP)
			split[i++] = PP;
		if(r & PPI)
			split[i++] = PPI;
		if(r & EQ)
			split[i++] = EQ;
		#elif defined (RCC8)
		if(r & PO)
			split[i++] = PO;
		if(r & EC)
			split[i++] = EC;
		if(r & DC)
			split[i++] = DC;
		if(r & TPP)
			split[i++] = TPP;
		if(r & NTPP)
			split[i++] = NTPP;
		if(r & TPPI)
			split[i++] = TPPI;
		if(r & NTPPI)
			split[i++] = NTPPI;
		if(r & EQ)
			split[i++] = EQ;
		#endif
	}
}
